//	ContainerWindow.cpp

#include <Debug.h>

#include "ContainerWindow.h"
#include "PluginEditors.h"
#include "PluginView.h"
#include <StLocker.h>
#include "EditorPanel.h"
#include "PluginProtocol.h"
#include "DataStreams.h"
#include "AsyncAlert.h"
#include "DataStreams.h"



enum {
	EDITOR_CHOSEN =			'echo',
	CLOSE_REPLY =			'clos',
	REALLY_REVERT =			'rrev'
};


ContainerWindow::ContainerWindow(
	BRect			area,
	const char *	title,
	window_type		type,
	ulong			flags) :
	BWindow(area, title, type, flags)
{
	fFileMenu = NULL;
	fTheView = NULL;
	fMenuBar = NULL;
	fDirty = false;
	fCloseAlert = NULL;
	fRevertAlert = NULL;
	fHasFile = false;
}


ContainerWindow::~ContainerWindow()
{
	RemoveLostMenus();
	if (fDocMIME) {
		free(fDocMIME);
	}
}


void
ContainerWindow::SetupWindow()
{
	StLocker<BWindow> lock(this);

	BRect area = Bounds();
	fFileMenu = MakeFileMenu();
	ASSERT(fFileMenu != NULL);
	fMenuBar = new BMenuBar(area, "main");
	fMenuBar->AddItem(fFileMenu);
	AddChild(fMenuBar);
}


BMenu *
ContainerWindow::MakeFileMenu()
{
	BMenu *fileMenu = new BMenu("File");

	BMenuItem *mClose = new BMenuItem("Close", new BMessage(B_QUIT_REQUESTED), 'W');
	BMenuItem *mSave = new BMenuItem("Save", new BMessage(CMD_SAVE), 'S');
	BMenuItem *mSaveAs = new BMenuItem("Save As", new BMessage(CMD_SAVE_AS), 'S',
		B_COMMAND_KEY | B_SHIFT_KEY);
	BMenuItem *mRevert = new BMenuItem("Revert", new BMessage(CMD_REVERT));
	BMenuItem *mPrint = new BMenuItem("Print", new BMessage(CMD_PRINT), 'P');

	fileMenu->AddItem(mClose);
	fileMenu->AddItem(mSave);
	fileMenu->AddItem(mSaveAs);
	fileMenu->AddItem(mRevert);
	fileMenu->AddSeparatorItem();
	fileMenu->AddItem(mPrint);

	return fileMenu;
}


long
ContainerWindow::OpenFile(
	record_ref		ref,
	ulong			type,
	const char *	mime)
{
	StLocker<BWindow> lock(this);

	if (fTheView) {
		RemoveChild(fTheView);
		delete fTheView;
		fTheView = NULL;
	}
	if (fDocMIME) {
		free(fDocMIME);
		fDocMIME = NULL;
	}

	BRect bnd = Bounds();
	bnd.top = fMenuBar->Bounds().bottom+1.0;
	if (bnd.top < 16.0) {
		printf("bnd.top is %f\n", bnd.top);
		bnd.top = 16.0;
	}

	EDITID editor = 0;
	long err = EDITFindEditor(TRUE, type, mime, editor);

	if (!err) {
		err = EDITMakePlugin(
			editor, 		//	editor
			bnd, 			//	area
			"editor", 		//	name
			B_FOLLOW_ALL, 	//	resize
			B_WILL_DRAW, 	//	flags
			this, 			//	container
			NULL,			//	extension
			fTheView);		//	outView
	}
	if (err)
		return err;

	ASSERT(fTheView != NULL);
	BFile *aFile = new BFile;
	err = aFile->SetRef(ref);
	if (err) {
		delete fTheView;
		delete aFile;
		return err;
	}
	char name[B_FILE_NAME_LENGTH];
	if (!aFile->GetName(name)) {
		SetTitle(name);
	}
	DFileStream *strm = new DFileStream(aFile);

	fDocType = type;
	fDocMIME = mime ? strdup(mime) : NULL;
	fHasFile = TRUE;
	fFileRef = ref;
	AddChild(fTheView);
	err = fTheView->Input(strm, fDocType, fDocMIME);

	Activate();
	Show();

	return err;
}


long
ContainerWindow::NewFile()
{
	StLocker<BWindow> lock(this);
	char str[200];
	strcpy(str, Title());
	strcat(str, " - New editor");
	BMessage *message = new BMessage(EDITOR_CHOSEN);
	if (!fEditorPanel) {
		EditorPanel *panel = new EditorPanel(str, message, this);
		fEditorPanel = panel;
	}
	fEditorPanel->Go();
	return B_NO_ERROR;
}


void
ContainerWindow::MessageReceived(
	BMessage *		message)
{
	switch (message->what) {

	//	plugin protocol messages

	case VIEW_FOCUS_MENUS:
		DoFocusMenus(
			message,
			dynamic_cast<PluginView *>(message->FindObject("view")));
		break;
	case VIEW_LOST_FOCUS:
		DoLostFocus(
			message,
			dynamic_cast<PluginView *>(message->FindObject("view")));
		break;
	case VIEW_DONE_EDITTING:
		DoDoneEditting(
			dynamic_cast<PluginView *>(message->FindObject("view")),
			message->FindBool("save"));
		break;
	case VIEW_PLEASE_SAVE:
		DoPleaseSave(
			dynamic_cast<PluginView *>(message->FindObject("view")));
		break;
	case VIEW_DATA_CHANGED:
		DoDataChanged(
			dynamic_cast<PluginView *>(message->FindObject("view")));
		break;
	case VIEW_SELECT_FRAME:
		DoSelectFrame(
			dynamic_cast<PluginView *>(message->FindObject("view")));
		break;

	//	other notification messages

	case EDITOR_CHOSEN:
		DoEditorChosen(message);
		break;
	case CLOSE_REPLY:
		DoCloseReply(message);
		break;
	case REALLY_REVERT:
		DoReallyRevert(message);
		break;

	//	regular commands

	case CMD_SAVE:
		DoSave(message);
		break;
	case CMD_SAVE_AS:
		DoSaveAs(message);
		break;
	case CMD_REVERT:
		DoRevert(message);
		break;
	case CMD_PRINT:
		DoPrint(message);
		break;

	//	otherwise

	default:
		inherited::MessageReceived(message);
		break;
	}
}


bool
ContainerWindow::QuitRequested()
{
	if (fRevertAlert) {
		fRevertAlert->Quit();
		fRevertAlert = NULL;
	}
	if (!fTheView) {
		return TRUE;
	}
	if (!fDirty)
		return true;
	if (fCloseAlert) {
		fCloseAlert->Go();
		return false;
	}
	char str[200];
	sprintf(str, "Do you want to save changes to '%s' before closing?", Title());
	fCloseAlert = new AsyncAlert(
			str, 
			new BMessage(CLOSE_REPLY), this, 
			"Save", "Cancel", "Close");
	fCloseAlert->Go();
	return false;
}


void
ContainerWindow::DoCloseReply(
	BMessage *message)
{
	long a = message->FindLong("answer");

	switch (a) {
	case 0:
	case 2:
		break;
	case 3:
		Quit();
		break;
	default:
		DoSave(message);
		break;
	}
}


void
ContainerWindow::MenusWillShow()
{
	BMenuItem *item = fFileMenu->FindItem(CMD_SAVE);
	item->SetEnabled(fDirty && fTheView);
	item = fFileMenu->FindItem(CMD_REVERT);
	item->SetEnabled(fDirty && fTheView && fHasFile);
	item = fFileMenu->FindItem(CMD_SAVE_AS);
	item->SetEnabled(fTheView != NULL);
	item = fFileMenu->FindItem(CMD_PRINT);
	item->SetEnabled(fTheView != NULL);
	if (fTheView)
		fTheView->MenusWillShow();
}


void
ContainerWindow::Activate(
	bool toActive)
{
	inherited::Activate(toActive);
	if (toActive)
		fTheView->MakeFocus(TRUE);
}


void
ContainerWindow::DoFocusMenus(
	BMessage *		message,
	PluginView *	forView)
{
	RemoveLostMenus();
	ASSERT(forView == fTheView);
	long count;
	ulong type;
	message->GetInfo("menus", &type, &count);
	ASSERT(type == B_OBJECT_TYPE);
	for (int ix=0; ix<count; ix++) {
		BMenu *menu = dynamic_cast<BMenu *>(message->FindObject("menus", ix));
		ASSERT(menu != NULL);
		fMenuBar->AddItem(menu);
		if (!fLostMenus.HasItem(menu)) {
			fLostMenus.AddItem(menu);
		}
	}
}


void
ContainerWindow::RemoveLostMenus()
{
	while (fLostMenus.CountItems() > 0) {
		BMenu *menu = fLostMenus.RemoveItemIndex(0);
		fMenuBar->RemoveItem(menu);
	}
}


void
ContainerWindow::DoLostFocus(
	BMessage *		message,
	PluginView *	forView)
{
	ASSERT(forView == fTheView);
	long count;
	ulong type;
	message->GetInfo("menus", &type, &count);
	ASSERT(type == B_OBJECT_TYPE);
	for (int ix=0; ix<count; ix++) {
		BMenu *menu = dynamic_cast<BMenu *>(message->FindObject("menus", ix));
		ASSERT(menu != NULL);
		if (!fLostMenus.HasItem(menu)) {
			fLostMenus.AddItem(menu);
		}
	}
}


void
ContainerWindow::DoDoneEditting(
	PluginView *	view,
	bool			save)
{
	ASSERT(view == fTheView);
	if (!save) {
		fDirty = false;
	}
	PostMessage(B_QUIT_REQUESTED);
}


void
ContainerWindow::DoPleaseSave(
	PluginView *	view)
{
	ASSERT(view == fTheView);
	PostMessage(CMD_SAVE);
}


void
ContainerWindow::DoDataChanged(
	PluginView *	view)
{
	ASSERT(view == fTheView);
	fDirty = true;
}


void
ContainerWindow::DoSelectFrame(
	PluginView *	view)
{
	ASSERT(view == fTheView);
	//	Do nothing
	//	A window doesn't select the frame of a view, because it doesn't 
	//	change the view's frame explicitly; only when itself gets changed.
	//	A container view that embedded other views, however, should be 
	//	prepared to deal with this message in order for the user to be 
	//	able to manipulate the view. While the view's frame is selected, 
	//	the view's outline should be drawn by the embedding view.
}


void
ContainerWindow::DoPrint(
	BMessage *		message)
{
	if (!fTheView)
		return;
	char title[200];
	strcpy(title, Title());
	BPrintJob *job = new BPrintJob(title);
	long err = job->InitJob();
	if (err || !job->CanContinue()) goto print_error;
	BRect paper = job->PrintableRect();
	paper.OffsetTo(B_ORIGIN);
	long numPages = fTheView->GetNumPages(paper);
	for (int ix=0; ix<numPages; ix++) {
		if (!job->CanContinue())
			goto print_error;
		BRect draw = fTheView->GetPageRect(ix, paper);
		job->DrawView(fTheView, draw, B_ORIGIN);
		job->SpoolPage();
	}
	if (!job->CanContinue())
		goto print_error;
	job->Commit();
	delete job;
	return;

print_error:
	job->CancelJob();
	delete job;
	if (err) {	//	not user cancel
		char str[200];
		sprintf(str, "Printing of '%s' could not be completed.\n\nError: %08x",
			Title(), err);
		AsyncAlert *alert = new AsyncAlert(str);
		alert->Go();
	}
}


void
ContainerWindow::DoSave(
	BMessage *		message)
{
	ASSERT(fTheView);
	if (!fTheView)
		return;

	if (!fHasFile) {
		DoSaveAs(message);
		return;
	}
	//	just re-use the previous stream
	long err = fTheView->Output(NULL, fDocType, fDocMIME);
	if (err) {
		char str[200];
		sprintf(str, "Could not save '%s'\n\nError: %08x",
			Title(), err);
		AsyncAlert *alrt = new AsyncAlert(str);
		alrt->Go();
	} else {
		fDirty = FALSE;
	}
}


void
ContainerWindow::DoSaveAs(
	BMessage *		message)
{
	RunSavePanel(Title(), "Save document");
}


void
ContainerWindow::DoRevert(
	BMessage *		message)
{
	if (!fDirty || !fHasFile)
		return;
	if (fRevertAlert) {
		fRevertAlert->Go();
		return;
	}
	char str[200];
	sprintf(str, "Really revert '%s'? Unsaved changes will be lost.",
		Title());
	fRevertAlert = new AsyncAlert(
			str, 
			new BMessage(REALLY_REVERT), this, 
			 "Revert", "Cancel");
	fRevertAlert->Go();
}


void
ContainerWindow::DoEditorChosen(
	BMessage *message)
{
	fEditorPanel = NULL;
	EDITID editor = message->FindLong("editor");
	if (editor < 0) {
		PostMessage(B_QUIT_REQUESTED);
		return;
	}
	BRect bnd = Bounds();
	bnd.top = fMenuBar->Bounds().bottom+1;
	long err = EDITMakePlugin(
			editor, 		//	editor
			bnd, 			//	area
			"editor", 		//	name
			B_FOLLOW_ALL, 	//	resize
			B_WILL_DRAW, 	//	flags
			this, 			//	container
			NULL,			//	extension
			fTheView);		//	outView
	if (err) {
		char str[200];
		sprintf(str, "The editor you chose cannot be used.\n\nError: %08x",
			err);
		AsyncAlert *alert = new AsyncAlert(str);
		alert->Go();
		PostMessage(B_QUIT_REQUESTED);
		return;
	}
	AddChild(fTheView);
	Activate();
	Show();
}


void
ContainerWindow::DoReallyRevert(
	BMessage *message)
{
	fRevertAlert = NULL;
	if (!fHasFile)
		return;
	if (message->FindLong("answer") != 1) {
		return;
	}
	ASSERT(fHasFile);
	ASSERT(fTheView);
	BFile *aFile = new BFile;
	long err = aFile->SetRef(fFileRef);
	if (!err) {
		DFileStream *strm = new DFileStream(aFile);
		err = fTheView->Input(strm, fDocType, fDocMIME);
	}
	if (err) {
		char str[200];
		sprintf(str, "Could not revert '%s'\n\nError: %08x",
			Title(), err);
		AsyncAlert *alrt = new AsyncAlert(str);
		alrt->Go();
	} else {
		fDirty = FALSE;
	}
}


void
ContainerWindow::SaveRequested(
	record_ref directory,
	const char *name)
{
	BFile *file = new BFile;
	BDirectory dir(directory);
	long err;
	if (0 != (err = dir.Create(name, file))) {
		if (0 != (err = dir.GetFile(name, file))) {
error:
			char str[200];
			sprintf(str, "Could not save to file '%s'.\n\nError: %08x",
				name, err);
			AsyncAlert *alert = new AsyncAlert(str);
			alert->Go();
			return;
		}
	}
	SetTitle(name);
	fFileRef = file->Record()->Ref();
	fHasFile = TRUE;
	DFileStream *strm = new DFileStream(file);
	const char *mime = NULL;
	ulong type = kAnyType;
	err = fTheView->Output(strm, type, mime);
	if (err) {
		goto error;
	}
	fDirty = FALSE;
}
