//	PluginEditors.cpp
//
//	Implementation of PluginEditors library functions to 
//	manage the user's set of plugin editors

#include "PluginEditors.h"
#include "PluginView.h"
#include "PluginProtocol.h"
#include "StAcquire.h"
#include "Datatypes.h"
#include <MList.h>


struct EVInfo {
					EVInfo() { memset(this, 0, sizeof(*this)); }

	EVInfo *		next;

	EDITID			pluginID;
	record_ref		ref;
	image_id		image;
	const char *	pluginName;
	const char *	pluginInfo;
	long			pluginVersion;
	const EDITInfo *pluginFormats;

	long			(*MakePlugin)(
						const BRect &	area,
						const char *	name,
						ulong			resize,
						ulong			flags,
						BHandler *		target,
						BMessage *		extension,
						PluginView * &	view);

};


static EVInfo *	sInfo;
static sem_id	sSem;


const char editorDefaultEditorPath[]		= "/boot/Editors";
const char editorEditorFolders[]			= "EditorFolders";
const char editorPreferredTableName[]		= "PreferredEditor";
const char editorPreferredTypeKeyName[]		= "Type";
const char editorPreferredMIMEKeyName[]		= "MIME";
const char editorPreferredEditorName[]		= "Editor";
const char editorForEditingKeyName[]		= "Editing";

const char *
EDITVersion(	//	returns version string
	long & outCurVersion,	//	current version spoken
	long & outMinVersion)	//	minimum version understood
{
	static char vString[] = "PluginEditors library v1.0 " __DATE__;
	outCurVersion = EDIT_CURRENT_VERSION;
	outMinVersion = EDIT_MIN_VERSION;
	return vString;
}


static long
LoadEditor(
	record_ref ref,
	EDITID & outEditor)
{
	//	check that this ref is not already loaded
	bool found = false;
	for (EVInfo *i = sInfo; i && !found; i=i->next)
		if (i->ref == ref) {
			found = true;
			outEditor = i->pluginID;
		}
	if (found)
		return B_NO_ERROR;

	//	sanity-check the file
	BFile file(ref);
	long err;
	if (0 != (err = file.Error()))
		return err;

	ulong type, app;
	if (0 != (err = file.GetTypeAndApp(&type, &app)))
		return err;

	if ((type != 'BAPP') && (type != EDITOR_TYPE))
		return EDIT_NO_EDITOR;

	//	go ahead and load it
	image_id image = load_add_on(&file);
	if (image < 0)
		return image;

	EVInfo *info = new EVInfo;
	info->ref = ref;
	info->image = image;
	info->pluginID = image^EDIT_CURRENT_VERSION;

	//	find all the symbols
	err = get_image_symbol(image, "pluginName", B_SYMBOL_TYPE_DATA,
			(void **)&info->pluginName);
	if (!err)
		err = get_image_symbol(image, "pluginInfo",
				B_SYMBOL_TYPE_DATA, (void **)&info->pluginInfo);
	if (!err)
		err = get_image_symbol(image, "pluginVersion",
				B_SYMBOL_TYPE_DATA, (void **)&info->pluginVersion);
	if (!err)
		err = get_image_symbol(image, "pluginFormats",
				B_SYMBOL_TYPE_DATA, (void **)&info->pluginFormats);

	if (!err)
		err = get_image_symbol(image, "MakePlugin", B_SYMBOL_TYPE_TEXT,
				(void **)&info->MakePlugin);

	if (err) {	//	this is not a correct add-on
		unload_add_on(image);
		delete info;
		return err;
	}

#if DEBUG
	{
		char name[B_FILE_NAME_LENGTH];
		file.GetName(name);
		puts(name);
	}
#endif

	//	add to global list
	info->next = sInfo;
	sInfo = info;
	outEditor = info->pluginID;

	return B_NO_ERROR;
}


static void
LoadDir(
	record_ref dirref,
	long & loadErr,
	long & nLoaded)
{
	long err;

	BDirectory dir(dirref);
	if (B_NO_ERROR != (err = dir.Error())) {

		loadErr = err;
		return;
	}

	int fileCount = dir.CountFiles();
	if (fileCount > 0) {
		record_ref *refs = new record_ref[fileCount];
		err = dir.GetFiles(0, fileCount, refs);
		if (err)
			loadErr = err;
		else for (int ix=0; ix<fileCount; ix++) {
			EDITID id;
			err = LoadEditor(refs[ix], id);
			if (err)
				loadErr = err;
			else
				nLoaded++;
		}
		delete[] refs;
	}
}


long
EDITInit(		//	establish connection
	ulong /* mySignature */,
	BQuery * editors)
{
	if (sSem <= 0)
		sSem = create_sem(1, "PluginEditors Lock");
	if (sSem <= 0)
		return sSem;

	StAcquire lock(sSem);

	long loadErr = 0;
	long nLoaded = 0;

	if (editors) {

		//	build a list of records that we want to load
		database_id db;
		MList<record_id> records;

		db = editors->Database()->ID();
		if (db < 0)
			return db;	//	error !

		long max = editors->CountRecordIDs();
		for (int ix=0; ix<max; ix++) {

			record_id id = editors->RecordIDAt(ix);
			if (id < 0)
				loadErr = id;	//	error!
			else
				records.AddItem(id);
		}

		//	load from one database
		for (int ix=0; ix<records.CountItems(); ix++) {
	
			//	extract the item
			record_ref ref;
			ref.database = db;
			ref.record = records.ItemAt(ix);
	
			EDITID id;
			long err = LoadEditor(ref, id);
			if (err)
				loadErr = err;
			else
				nLoaded++;
		}
	} else {

		//	load whatever the defaults are

		//	get preferences from database
		long size = 0;
		record_ref *dref = NULL;
		long err = DATAGetSetting(editorEditorFolders, (void *&)dref, size);

		size = size / sizeof(record_ref);

		//	if we have nothing to load, just cop out
		if (!err && size) {

			//	for each folder record_ref in the list, load its files
			for (int iy=0; iy<size; iy++) {
				//	we don't remove directories that are unavailable, since 
				//	they may reference a volume that's just currently unmounted.

				LoadDir(dref[iy], loadErr, nLoaded);
			}
		}
		//	always try the default path
		const char *dtpath = getenv("EDITORS");
		if (!dtpath)
			dtpath = editorDefaultEditorPath;
		record_ref eref;
		err = get_ref_for_path(dtpath, &eref);
		if (!err) {
			LoadDir(eref, loadErr, nLoaded);
		}

		free(dref);
	}

	if (nLoaded)
		loadErr = B_NO_ERROR;
	return loadErr;
}


long
EDITShutdown()	//	don't want to talk anymore
{
	if (sSem > 0) {

		acquire_sem(sSem);

		EVInfo *info = sInfo;
		sInfo = NULL;
		while (info != NULL) {
			EVInfo *del = info;
			info = info->next;
			unload_add_on(del->image);
			delete del;
		}

		delete_sem(sSem);
		sSem = 0;

	} else {

		return EDIT_NOT_INITIALIZED;	
	}
	return B_NO_ERROR;
}


long
EDITFindEditor(	//	find out what can edit something
	bool				inForEditing,
	ulong				inDataType,//	kAnyType for MIME 
	const char *		inMIMEType,
	EDITID &			outEditor)
{
	if (!sSem)
		return EDIT_NOT_INITIALIZED;

	printf("EDITFindEditor(%d, %.4s,%s)\n", inForEditing, &inDataType, inMIMEType?inMIMEType:"(null)");

	long err = EDITGetPreferredEditor(inForEditing, inDataType, inMIMEType, outEditor);
	EVInfo *info = sInfo;
	if (err) {
		while (info) {
			for (int ix=0; info->pluginFormats[ix].formatName[0]; ix++) {
				printf("%d: %x %.4s %s\n",
					info->pluginID,
					info->pluginFormats[ix].formatFlags,
					&(info->pluginFormats[ix].formatType),
					info->pluginFormats[ix].MIMEName);
				if (inForEditing & !(info->pluginFormats[ix].formatFlags & kCanEdit))
					continue;	//	next format in the for loop
				if (inDataType &&
						(info->pluginFormats[ix].formatType == inDataType)) {
					printf("matched data type\n");
					goto match;
				}
				if (inMIMEType && !strcmp(
						info->pluginFormats[ix].MIMEName, inMIMEType)) {
					printf("matched MIME type\n");
					goto match;
				}
			}
			info = info->next;
		}
match:
		if (!info) {
			printf("No editor.\n");
			return EDIT_NO_EDITOR;
		}
		outEditor = info->pluginID;
	}
	printf("Got editor %d\n", outEditor);
	return B_NO_ERROR;
}


long
EDITEditorFromFile(//	find editor associated with file
	record_ref			inFile,
	EDITID &			outEditor,
	bool				inTryToLoad)//	try to load if not found
{
	if (!sSem)
		return EDIT_NOT_INITIALIZED;

	EVInfo *info = sInfo;
	while (info) {
		if (info->ref == inFile) {
			return info->pluginID;
		}
		info = info->next;
	}
	if (!inTryToLoad) {
		return EDIT_NO_EDITOR;
	}
	StAcquire lock(sSem);
	return LoadEditor(inFile, outEditor);
}


long
EDITGetAllEditors(//	find all handler IDs
	EDITID * &			outList,//	call delete[] when done
	long &				outCount)
{
	if (!sSem)
		return EDIT_NOT_INITIALIZED;

	EVInfo *info = sInfo;
	long editorCount = 0;
	while (info) {
		info = info->next;
		editorCount++;
	}
	outList = new EDITID[editorCount];
	outCount = editorCount;
	info = sInfo;
	int ix = 0;
	while (info) {
		outList[ix] = info->pluginID;
		ix++;
		info = info->next;
	}
	return B_NO_ERROR;
}


long
EDITGetEditorFormats(//	find formats for an editor
	EDITID				inEditor,
	const EDITInfo * &	outFormats,//	don't write contents!
	long &				outNumFormats)
{
	if (!sSem)
		return EDIT_NOT_INITIALIZED;
	EVInfo *info = sInfo;
	while (info) {
		if (info->pluginID == inEditor) {
			outFormats = info->pluginFormats;
			int ix;
			for (ix=0; info->pluginFormats[ix].formatName[0]; ix++)
				;	//	just count
			outNumFormats = ix;
			return B_NO_ERROR;
		}
		info = info->next;
	}
	return EDIT_NO_EDITOR;
}


long
EDITGetEditorInfo(
	EDITID				inEditor,
	const char * &		outName,
	const char * &		outInfo,
	ulong &				outVersion)
{
	if (!sSem)
		return EDIT_NOT_INITIALIZED;
	EVInfo *info = sInfo;
	while (info) {
		if (info->pluginID == inEditor) {
			outName = info->pluginName;
			outInfo = info->pluginInfo;
			outVersion = info->pluginVersion;
			return B_NO_ERROR;
		}
		info = info->next;
	}
	return EDIT_NO_EDITOR;
}


long
EDITMakePlugin(	//	create PluginView subclass
	EDITID				inEditor,
	const BRect &		inArea,
	const char *		inName,
	ulong				inResize,
	ulong				inFlags,
	BHandler *			inContainer,
	BMessage *			inExtension,
	PluginView * &		outView)
{
	outView = NULL;
	if (!sSem)
		return EDIT_NOT_INITIALIZED;
	if (!inName)
		return EDIT_BAD_PARAMETERS;
	if (!inContainer)
		return EDIT_BAD_PARAMETERS;

	printf("EDITMakePlugin(%d)\n", inEditor);

	EVInfo *info = sInfo;
	while (info) {
		if (info->pluginID == inEditor) {
			break;
		}
		info = info->next;
	}
	if (!info)
		return EDIT_NO_EDITOR;

	return info->MakePlugin(inArea, inName, inResize, inFlags,
			inContainer, inExtension, outView);
}


static BTable *
CreateTable(
	BDatabase & database)
{
	BTable *ret = database.CreateTable(editorPreferredTableName);
	if (!ret)
		return NULL;
	ret->AddLongField(editorPreferredTypeKeyName, B_INDEXED_FIELD);
	ret->AddStringField(editorPreferredMIMEKeyName, B_INDEXED_FIELD);
	ret->AddStringField(editorPreferredEditorName);
	ret->AddLongField(editorForEditingKeyName);
	return ret;
}


long
EDITSetPreferredEditor(
	EDITID				inEditor,
	bool				inForEditing,
	ulong				inForType,
	const char *		inForMIME)
{
	if (!sSem)
		return EDIT_NOT_INITIALIZED;
	if (!inForType && !inForMIME)
		return EDIT_BAD_PARAMETERS;

	const char * editorName;
	const char * editorInfo;
	ulong editorVersion;
	long err = EDITGetEditorInfo(inEditor, editorName, editorInfo, editorVersion);
	if (err)
		return err;

	BVolume vol = boot_volume();
	BTable *table = vol.Database()->FindTable(editorPreferredTableName);
	if (!table)
		table = CreateTable(*vol.Database());
	if (!table)
		return B_STORAGE_ERROR_BASE;	//	why is there no better error for this?
	BQuery *q = new BQuery;
	q->AddTable(table);
	if (inForType) {
		q->PushLong(inForType);
		q->PushField(editorPreferredTypeKeyName);
	} else {
		q->PushString(inForMIME);
		q->PushField(editorPreferredMIMEKeyName);
	}
	q->PushOp(B_EQ);
	q->PushLong(inForEditing);
	q->PushField(editorForEditingKeyName);
	q->PushOp(B_EQ);
	q->PushOp(B_AND);
	q->FetchOne();
	BRecord *theRec = NULL;
	if (!q->CountRecordIDs()) {
		theRec = new BRecord(table);
		theRec->Lock();
		const char * mimeStr = (inForMIME ? inForMIME : "");
		theRec->SetString(editorPreferredMIMEKeyName, mimeStr);
		theRec->SetLong(editorPreferredTypeKeyName, inForType);
		theRec->SetLong(editorForEditingKeyName, inForEditing);
	} else {
		theRec = new BRecord(vol.Database(), q->RecordIDAt(0));
		theRec->Lock();
	}
	delete q;
	theRec->SetString(editorPreferredEditorName, editorName);
	err = theRec->Commit();
	theRec->Unlock();
	delete theRec;

	return err;
}


long
EDITGetPreferredEditor(
	bool				inForEditing,
	ulong				inForType,
	const char *		inForMIME,
	EDITID &			outEditor)
{
	if (!sSem)
		return EDIT_NOT_INITIALIZED;
	if (!inForType && !inForMIME)
		return EDIT_BAD_PARAMETERS;

	outEditor = 0;
	BVolume vol = boot_volume();
	BTable *table = vol.Database()->FindTable(editorPreferredTableName);
	if (!table)
		table = CreateTable(*vol.Database());
	if (!table)
		return B_STORAGE_ERROR_BASE;	//	why is there no better error for this?
	BQuery *q = new BQuery;
	q->AddTable(table);
	if (inForType) {
		q->PushLong(inForType);
		q->PushField(editorPreferredTypeKeyName);
	} else {
		q->PushString(inForMIME);
		q->PushField(editorPreferredMIMEKeyName);
	}
	q->PushOp(B_EQ);
	q->PushLong(inForEditing);
	q->PushField(editorForEditingKeyName);
	q->PushOp(B_EQ);
	q->PushOp(B_AND);
	q->FetchOne();
	BRecord *theRec = NULL;
	if (!q->CountRecordIDs()) {
		delete q;
		return B_NAME_NOT_FOUND;
	}
	BRecord rec(vol.Database(), q->RecordIDAt(0));
	delete q;
	const char *editorName = rec.FindString(editorPreferredEditorName);
	if (!editorName)
		return EDIT_NO_EDITOR;
	EVInfo *info = sInfo;
	while (info) {
		if (!strcmp(info->pluginName, editorName)) {
			outEditor = info->pluginID;
			return B_NO_ERROR;
		}
		info = info->next;
	}
	return EDIT_NO_EDITOR;
}


