//	PluginView.h
//
//	Base class for all lightweight content editors
//
//	Jon Watte (h+@metrowerks.com) 960926

#pragma once

#include <View.h>

class DStream;


//	This is the PluginView class that you should return an instance 
//	of in MakePlugin().
//	There are some operations a view should perform upon focus aquisition:
//	*	Send a FocusMenus message to its container with a list of the 
//		menus it wants in the menu bar.
//	*	Show or enable its scroll bars if needed.
//	There are soem operations a view should perform upon focus loss:
//	*	Hide or disable its scroll bars if needed.


#pragma export on



enum EDITInfoFlags {
	kNullFlags = 0,
	kCanEdit = 1
};

struct EDITInfo {
	ulong				formatType;				//	B_ASCII_TYPE, ...
	EDITInfoFlags		formatFlags;			//	kCanEdit, ...
	char				formatName[B_FILE_NAME_LENGTH];
	char				MIMEName[B_FILE_NAME_LENGTH];
};


#define EDITOR_TYPE		'edit'


class PluginView :
	public BView
{
public:
		//	The view is a regular view, but it also has a containing 
		//	view which it tells about changes in state (focus etc)

								PluginView(
									BRect			area, 
									const char *	name,
									ulong			resize, 
									ulong			flags, 
									BHandler *		container,
									BMessage *		extension);

		//	Typically, you want to delete any DStream you might be  
		//	hanging on to from Input() or Output().

								~PluginView();

		//	Used wants us to edit/not edit this thing
		//	return EDIT_CANNOT_EDIT if you're only a viewer (that's a 
		//	usage error since viewers will not be returned for a 
		//	request with canEdit = true)
		//	All editors should be able to go into non-editting mode

virtual	long					SetCanEdit(
									bool			inCanEdit) = 0;

		//	Regular overrides

		void					MessageReceived(
									BMessage *		message);


		//	These are public extensions to the BView base class


		//	DispatchMessage default case calls ExtensionMessage. 
		//	However, direct calls of ExtensionMessage (not through 
		//	the looper system) are also supported.
		//	Known messages should return TRUE, unknown/unhandled 
		//	should return FALSE.

virtual	bool					ExtensionMessage(
									BMessage *		inMessage,
									BMessage **		outMessage);

		//	Read data from wherever into the view. That data should 
		//	replace any data currently in the view. The stream is 
		//	"donated" to you; it's up to you to dispose of it when 
		//	you feel like it (but see Output).
		//	Pure virtual because you have to implement it.

virtual	long					Input(
									DStream *		inStream,
									ulong			inType,
									const char *	inMIMEType) = 0;

		//	Save the data in the view to wherever. TypeQualification 
		//	specifies what type to use like so:
		//	kAnyType (from Datatypes.lib) - save as your best type, 
		//	and set ioTypeQualification and ioMIMEType to what you used.
		//	kDefaultType (from Datatypes.lib) - save as the default 
		//	type for your content class; e g ASCII TEXT, and return what 
		//	you used.
		//	Any other value OR a non-NULL ioMIMEType - use the specified 
		//	type.
		//	Returns in ioMIMEType will not be disposed by caller, can be 
		//	const strings or pointer to member data.
		//	If outStream is NULL you should re-use the stream you got 
		//	from Input(). If you never got one, return an error. If 
		//	it is not NULL, you should dispose the old one and take 
		//	ownership of this new stream.
		//	Pure virtual because you have to implement it.

virtual	long					Output(
									DStream *		outStream,
									ulong &			ioTypeQualification,
									const char * &	ioMIMEType) = 0;

		//	This function is (synchronously) called when the menus are
		//	about to be shown and your view has the focus. You can 
		//	assume that any menus you define are in the menu bar. Fix 
		//	them up as you see fit.

virtual	void					MenusWillShow();

		//	For printing and frame space negotiation, these functions 
		//	can be overridden to return information needed by the 
		//	containing view/window.
		//	GetScreenArea() should return the preferred on-screen area. 
		//	GetImageArea() should return the total size of the image 
		//	that the view displays.
		//	GetNumPages() should return the total number of pages if 
		//	given the rect to print in for each page.
		//	GetPageRect() should return the rect to draw for page # 
		//	X where X starts at 0.
		//	If you hook into the print loop, you can make the following 
		//	assumptions:
		//	The print loop will call GetNumPages() once before starting 
		//	to paginate.
		//	It will then call GetPageRect() once right before printing 
		//	that same page.
		//	If you want to print page numbers or something like that, 
		//	you can get the page number from the last call to GetPageRect().

virtual	BRect					GetScreenArea();
virtual	BRect					GetImageArea();
virtual	long					GetNumPages(
									const BRect &	forArea);
virtual	BRect					GetPageRect(
									long pageIx,
									const BRect &	forArea);

protected:

		//	Data members that each plugin view has

		BHandler *				fContainer;


		//	Utility functions for view callbacks to container

		//	Just send the message
		long					PostToContainer(
									BMessage *		inMessage);

		//	We got focus - tell container about our menus.
		//	Menus should be passed as BList of BMenu objects.
		//	Views have to make arrangements to get their menu 
		//	commands on their own; typically by calling 
		//	SetTarget() on its menu items.
		//	Also call when losing focus with same menu list.
		long					FocusMenus(
									bool			gotFocus,
									const BList &	inMenus);

		//	The user indicated he's "done editting" the content
		//	Note that the container may decide on this without view 
		//	intervention.
		long					DoneEditting(
									bool			doSave);

		//	The user indicated he wants to "save" the data
		long					PleaseSave();

		//	Send this when the user starts editing data after 
		//	it's clean (saved or freshly loaded)
		long					DataChanged();

		//	If you have a "select frame" command that allows 
		//	the user to select your frame (to resize it etc) 
		//	you should call this function to accomplish it.
		//	This will make the view lose its focus.
		long					SelectFrame();

};


extern "C" {

	extern	char			pluginName[];		//	required, C string, ex "Jon's Sound"
	extern	char			pluginInfo[];		//	required, descriptive C string, ex "Jon's Sound Handler (shareware $5: hplus@ix.netcom.com) v1.00"
	extern	long			pluginVersion;		//	required, integer, ex 100

	extern	EDITInfo		pluginFormats[];	//	required, end with NULL entry

	extern	long			MakePlugin(
								const BRect &	inArea,
								const char *	inTitle,
								ulong			inResize,
								ulong			inFlags,
								BHandler *		inTarget,
								BMessage *		inExtension,
								PluginView * &	outView);

}


#pragma export reset
