//	TextEditor.cpp

#include "PluginView.h"
#include "PluginProtocol.h"

#include <DStream.h>
#include <MList.h>
#include <Menu.h>
#include <TextView.h>
#include <MenuItem.h>
#include <ScrollBar.h>
#include <stdlib.h>



class TextPlugin :
	public PluginView
{
	friend class XTextView;

public:
								TextPlugin(
									BRect			area, 
									const char *	name,
									ulong			resize, 
									ulong			flags, 
									BHandler *		container,
									BMessage *		extension);
								~TextPlugin();

		void					AttachedToWindow();
		void					DetachedFromWindow();

		long					SetCanEdit(
									bool			inCanEdit);

		void					MessageReceived(
									BMessage *		fMessage);

		long					Input(
									DStream *		inStream,
									ulong			inType,
									const char *	inMIMEType);

		long					Output(
									DStream *		outStream,
									ulong &			ioTypeQualification,
									const char * &	ioMIMEType);

		void					MenusWillShow();

		void					Dirty();

protected:

		BScrollBar *			fVertical;
		BScrollBar *			fHorizontal;
		MList<BMenu *>			fMenus;
		DStream *				fStream;
		BTextView *				fTextView;
		bool					fHasFocus;
		bool					fDirty;

		BRect					GetImageArea();

		void					DoFocus(
									bool			toFocus);

};


char			pluginName[] = "TextEditor";

char			pluginInfo[] = "Simple sample text editor v1.0";

long			pluginVersion = 100;

	ulong				formatType;				//	B_ASCII_TYPE, ...
	EDITInfoFlags		formatFlags;			//	kCanEdit, ...
	char				formatName[B_FILE_NAME_LENGTH];
	char				MIMEName[B_FILE_NAME_LENGTH];

EDITInfo		pluginFormats[] = {
	{	'TEXT',	kCanEdit,	"ASCII Text",	"text/ascii"	},
	{	0,	kNullFlags,	"",	""	}
};

long
MakePlugin(
	const BRect &	inArea,
	const char *	inTitle,
	ulong			inResize,
	ulong			inFlags,
	BHandler *		inTarget,
	BMessage *		inExtension,
	PluginView * &	outView)
{
	outView = new TextPlugin(inArea, inTitle, inResize, inFlags, inTarget, inExtension);
	return B_NO_ERROR;
}



class XTextView :
	public BTextView
{
public:
								XTextView(
									BRect area,
									const char *name,
									BRect txRect,
									ulong resize,
									ulong flags,
									TextPlugin *parent) :
								BTextView(
									area,
									name,
									txRect,
									resize,
									flags),
								fParent(parent) { }

		void					MakeFocus(
									bool focus)
								{
									if (!focus)
										fParent->DoFocus(FALSE);
									inherited::MakeFocus(focus);
									if (focus)
										fParent->DoFocus(TRUE);
								}

		void					KeyDown(
									ulong aKey)
								{
									switch (aKey) {
									case B_LEFT_ARROW:
									case B_RIGHT_ARROW:
									case B_UP_ARROW:
									case B_DOWN_ARROW:
									case B_HOME:
									case B_END:
									case B_PAGE_UP:
									case B_PAGE_DOWN:
										break;
									default:
										fParent->Dirty();
										break;
									}
									inherited::KeyDown(aKey);
								}

protected:

		TextPlugin *			fParent;
};


TextPlugin::TextPlugin(
	BRect			area, 
	const char *	name,
	ulong			resize, 
	ulong			flags, 
	BHandler *		container,
	BMessage *		extension) :
	PluginView(area, name, resize, flags, container, extension)
{
	fStream = NULL;
	fHasFocus = FALSE;
	fDirty = FALSE;

	area = Bounds();
	area.right -= B_V_SCROLL_BAR_WIDTH;
	area.bottom -= B_H_SCROLL_BAR_HEIGHT;
	BRect txArea = area;
	txArea.OffsetTo(B_ORIGIN);
	txArea.left += 3;

	//	we delegate actual work to this fellow

	fTextView = new XTextView(area, "text", txArea, B_FOLLOW_ALL, B_WILL_DRAW, this);
	AddChild(fTextView);

	BRect sb(area);
	sb.left = sb.right+1;
	sb.right = sb.left+B_V_SCROLL_BAR_WIDTH;
	sb.bottom++;
	BScrollBar *bar = new BScrollBar(sb, "vert", fTextView, 0, 0, B_VERTICAL);
	AddChild(bar);
	fVertical = bar;
	fVertical->Hide();

	sb = area;
	sb.top = sb.bottom+1;
	sb.bottom = sb.top+B_H_SCROLL_BAR_HEIGHT;
	sb.right++;
	bar = new BScrollBar(sb, "hor", fTextView, 0, 0, B_HORIZONTAL);
	AddChild(bar);
	fHorizontal = bar;
	fHorizontal->Hide();
}


void
TextPlugin::AttachedToWindow()
{
	BMenu *editMenu = new BMenu("Edit");

	BMenuItem *item = new BMenuItem("Undo", new BMessage(CMD_UNDO), 'Z');
	item->SetTarget(this);
	editMenu->AddItem(item);

	editMenu->AddSeparatorItem();

	item = new BMenuItem("Cut", new BMessage(B_CUT), 'X');
	item->SetTarget(this);
	editMenu->AddItem(item);

	item = new BMenuItem("Copy", new BMessage(B_COPY), 'C');
	item->SetTarget(this);
	editMenu->AddItem(item);

	item = new BMenuItem("Paste", new BMessage(B_PASTE), 'V');
	item->SetTarget(this);
	editMenu->AddItem(item);

	item = new BMenuItem("Clear", new BMessage(CMD_CLEAR));
	item->SetTarget(this);
	editMenu->AddItem(item);

	editMenu->AddSeparatorItem();

	item = new BMenuItem("Select All", new BMessage(CMD_SELECT_ALL), 'A');
	item->SetTarget(this);
	editMenu->AddItem(item);

	item = new BMenuItem("Select Frame", new BMessage(CMD_SELECT_FRAME));
	item->SetTarget(this);
	editMenu->AddItem(item);

	fMenus.AddItem(editMenu);
}


void
TextPlugin::DetachedFromWindow()
{
	delete fMenus.RemoveItemIndex(0);
}


TextPlugin::~TextPlugin()
{
	delete fStream;
	if (!fHasFocus) {
		while (fMenus.CountItems()) {
			delete fMenus.RemoveItemIndex(0);
		}
	}
}


long
TextPlugin::SetCanEdit(
	bool		inCanEdit)
{
	fTextView->MakeEditable(inCanEdit);
	return B_NO_ERROR;
}


long
TextPlugin::Input(
		DStream *		inStream,
		ulong			inType,
		const char *	inMIMEType)
{
	if (!inStream)
		inStream = fStream;
	else {
		delete fStream;
		fStream = inStream;
	}
	//	Calling Size() may make the stream block for some 
	//	indeterminate time, if it's a network connection or 
	//	something like that. If we want to work well as parts 
	//	of a web browser, we should read small chunks (2k or 
	//	so) and insert them, until we get an error (no more 
	//	data to read)
	long size = inStream->Size();
	if (size < 0)
		return size;
	void *ptr = malloc(size);
	long err = inStream->Read(ptr, size);
	if (err < 0) {
		free(ptr);
		return err;
	}
	fTextView->SetText((char *)ptr, size);
	free(ptr);
	fDirty = FALSE;
	return B_NO_ERROR;
}


long
TextPlugin::Output(
	DStream *		outStream,
	ulong &			ioTypeQualification,
	const char * &	ioMIMEType)
{
	//	For the case of a NULL stream we're supposed to re-use the 
	//	previous stream. It's up to us to re-set it to the beginning, 
	//	since we may optimize access and writing.
	if (!outStream) {
		outStream = fStream;
		outStream->SetSize(0);
	}
	//	In the case of a new stream, it's presumably given to us in 
	//	the state our container wants, so we just go ahead and write 
	//	to it.
	else {
		delete fStream;
		fStream = outStream;
	}
	outStream->SetSize(0);
	long err = outStream->Write(fTextView->Text(), fTextView->TextLength());
	if (err < 0)
		return err;
	fDirty = FALSE;
	return B_NO_ERROR;
}


void
TextPlugin::MenusWillShow()
{
	//	Here, we should enable the menu items depending on 
	//	selection state and clipboard.

	BMenu *editMenu = fMenus.ItemAt(0);
	long start = 0, end = 0;
	fTextView->GetSelection(&start, &end);
	long size = fTextView->TextLength();
	editMenu->FindItem(CMD_SELECT_ALL)->SetEnabled((start != 0) || (end != size));
	editMenu->FindItem(B_CUT)->SetEnabled(start != end);
	editMenu->FindItem(B_COPY)->SetEnabled(start != end);
	editMenu->FindItem(CMD_CLEAR)->SetEnabled(start != end);
	editMenu->FindItem(CMD_UNDO)->SetEnabled(FALSE);
	editMenu->FindItem(B_PASTE)->SetEnabled(TRUE);	//	always paste...
	editMenu->FindItem(CMD_SELECT_FRAME)->SetEnabled(TRUE);
}


void
TextPlugin::DoFocus(
	bool			toFocus)
{
	fHasFocus = toFocus;
	if (toFocus) {
		fVertical->Show();
		fHorizontal->Show();
	} else {
		fHorizontal->Hide();
		fVertical->Hide();
	}
	FocusMenus(toFocus, fMenus);
}


void
TextPlugin::Dirty()
{
	if (!fDirty) {
		fDirty = TRUE;
		DataChanged();
	}
}


BRect
TextPlugin::GetImageArea()
{
	return fTextView->TextRect();
}


void
TextPlugin::MessageReceived(
	BMessage *message)
{
	switch (message->what) {
	case CMD_UNDO:
		//	do nothing
		break;
	case B_CUT:
		be_clipboard->Lock();
		fTextView->Cut(be_clipboard);
		Dirty();
		be_clipboard->Unlock();
		break;
	case B_COPY:
		be_clipboard->Lock();
		fTextView->Copy(be_clipboard);
		be_clipboard->Unlock();
		break;
	case B_PASTE:
		be_clipboard->Lock();
		fTextView->Paste(be_clipboard);
		Dirty();
		be_clipboard->Unlock();
		break;
	case CMD_CLEAR:
		fTextView->Delete();
		break;
	case CMD_SELECT_ALL:
		fTextView->SelectAll();
		break;
	case CMD_SELECT_FRAME:
		SelectFrame();
		break;
	default:
		inherited::MessageReceived(message);
		break;
	}
}
